/*-----------------------------------------------------------------------------
 *
 * Enhydra JdbcS
 * The Initial Developer of the Original Code is Lutris Technologies Inc.
 * Portions created by Lutris are Copyright 1997-2000 Lutris Technologies Inc.
 * All Rights Reserved.
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 * www.enhydra.org/license/epl.html
 *
 * Software distributed under the License is distributed on an "ASIS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific language governing rights and limitations under the License.
 *
 * -----------------------------------------------------------------------------
 */
package org.enhydra.jdbc.standard;

import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.CallableStatement;
import java.util.Hashtable;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;
import javax.transaction.RollbackException;
import javax.transaction.SystemException;

public class StandardXAConnectionHandle extends StandardConnectionHandle {

    boolean		globalTransaction;	// true if a global transaction is in effect
    public TransactionManager	transactionManager;
    public Transaction		tx;
    public StandardXAConnection xacon;
    public boolean thisAutoCommit = true;

    /**
     * Constructor
     */
    public StandardXAConnectionHandle (StandardXAConnection pooledCon, Hashtable preparedStatementCache, int preparedStmtCacheSize, TransactionManager tm) {
        super (pooledCon, preparedStatementCache, preparedStmtCacheSize); // setup StandardXAConnectionHandle
        xacon = pooledCon;
        transactionManager = tm;
        log = pooledCon.dataSource.log;

        // This will have set the Connection to the current Connection.
        // However this might change if a global transaction gets selected.
    }

    public void setTransactionManager(TransactionManager tm) {
        this.transactionManager = tm;
    }

    synchronized public void close() throws SQLException {
        Transaction ttx = tx;
        // note: ttx is used instead of tx because super.close(), call end()
        // on StdXAConnection which call tx = null;
        super.close();
        log.debug("StandardXAConnectionHandle:close");
        log.debug("StandardXAConnectionHandle:close globalTransaction='"+globalTransaction+
                "' con.getAutoCommit='"+con.getAutoCommit()+"' ttx='"+ttx+"'");

        if ((!con.getAutoCommit()) && (ttx == null)){
            log.debug("StandardXAConnectionHandle:close rollback the connection");
            con.rollback();
        } else
            log.debug("StandardXAConnectionHandle:close do nothing else");
    }

    /**
     * Called by the StandardXADataSource when a global transaction
     * gets associated with this connection.
     */
    void setGlobalTransaction (boolean setting) throws SQLException {
        log.debug("StandardXAConnectionHandle:setGlobalTransaction gTransaction='"+setting+"'");
        globalTransaction = setting;			// set global flag
        con = pooledCon.getPhysicalConnection();	// get the real connection
        if (con == null)
            log.warn("StandardXAConnectionHandle:setGlobalTransaction con is null before setupPreparedStatementCache");
        else
            log.debug("StandardXAConnectionHandle:setGlobalTransaction con is *NOT* null before setupPreparedStatementCache");
        //setupPreparedStatementCache();
        super.setAutoCommit (!setting);			// commits must be done by transaction manager
    }

    public void setAutoCommit(boolean autoCommit) throws SQLException {
        if (globalTransaction)			// if taking part in a global transaction
            throw new SQLException ("StandardXAConnectionHandle:setAutoCommit This connection is part of a global transaction");
        super.setAutoCommit(autoCommit);
    }

    public void commit() throws SQLException {
        if (globalTransaction)			// if taking part in a global transaction
            throw new SQLException ("StandardXAConnectionHandle:commit:This connection is part of a global transaction");
        super.commit();
        //tx = null;
    }

    public void rollback() throws SQLException {
        if (globalTransaction)			// if taking part in a global transaction
            throw new SQLException ("StandardXAConnectionHandle:rollback:This connection is part of a global transaction");
        super.rollback();
        //tx = null;
    }

    synchronized PreparedStatement checkPreparedCache (String sql, int type, int concurrency, Object lookupKey) throws SQLException {
        PreparedStatement ret = null;				// the return value
        // NOTE - We include the Connection in the lookup key. This has no
        // effect here but is needed by StandardXAConnection where the the physical
        // Connection used can vary over time depending on the global transaction.
        if (preparedStatementCache != null) {
            Object obj = preparedStatementCache.get (lookupKey);// see if there's a PreparedStatement already
            if (obj != null) {					// if there is
                log.debug("StandardXAConnectionHandle:checkPreparedCache object is found");
                ret = (PreparedStatement)obj;		// use as return value
                try {
                    ret.clearParameters();			// make it look like new
                } catch (SQLException e) {
                    // Bad statement, so we have to create a new one
                    ret = createPreparedStatement(sql, type, concurrency); // create new prepared statement
                }
                preparedStatementCache.remove(lookupKey);	// make sure it cannot be re-used
                inUse.put (lookupKey, ret);			// make sure it gets reused by later delegates
            } else {					// no PreparedStatement ready
                log.debug("StandardXAConnectionHandle:checkPreparedCache object is *NOT* found");
                ret = createPreparedStatement(sql, type, concurrency); // create new prepared statement
                inUse.put (lookupKey, ret);			// will get saved in prepared statement cache
            }
        } else {
            log.debug("StandardXAConnectionHandle:checkPreparedCache object the cache is out");
            ret = createPreparedStatement(sql, type, concurrency); // create new prepared statement
        }
        // We don't actually give the application a real PreparedStatement. Instead
        // they get a StandardPreparedStatement that delegates everything except
        // PreparedStatement.close();
        log.debug("StandardXAConnectionHandle:checkPreparedCache pstmt='"+ret.toString()+"'");
        return ret;
    }

    /**
     * Creates a PreparedStatement for the given SQL. If possible, the
     * statement is fetched from the cache.
     */
    public PreparedStatement prepareStatement(String sql) throws SQLException {
        if (tx == null) {
            log.debug("StandardXAConnectionHandle:prepareStatement tx==null");
            try {
                try {
                    Transaction ntx = (transactionManager!=null)?transactionManager.getTransaction():null;
                    if (ntx != null) {
                        tx = ntx;
                        thisAutoCommit = this.getAutoCommit();
                        this.setAutoCommit(false);
                        try {
                            tx.enlistResource(xacon.getXAResource());// enlist the xaResource in the transaction
                        } catch (RollbackException n) {
                            n.printStackTrace();
                            throw new SQLException("StandardXAConnectionHandle:prepareStatemnet enlistResource exception");
                        }
                    }
                } catch (SystemException n) {
                    n.printStackTrace();
                    throw new SQLException("StandardXAConnectionHandle:prepareStatement getTransaction exception: "+n.toString());
                }
            } catch (NullPointerException n) {
                // current is null: we are not in EJBServer.
                n.printStackTrace();
                throw new SQLException("StandardXAConnectionHandle:prepareStatement should not be used outside an EJBServer");
            }
        } else
            log.debug("StandardXAConnectionHandle:prepareStatement tx!=null");

        // if you want to use a REAL PrepareStatement object, please
        // uncomment the 2 following lines and comment the last one.
        //PreparedStatement ops = con.prepareStatement(sql);
        //return ops;
        isReallyUsed = true;
        return new StandardXAPreparedStatement (this, sql, 0, 0);
    }

    /**
     * Creates a PreparedStatement for the given SQL, type and concurrency.
     * If possible, the statement is fetched from the cache.
     */
    public PreparedStatement prepareStatement(String sql,int resultSetType,int resultSetConcurrency) throws SQLException {
        return new StandardXAPreparedStatement (this, sql, resultSetType, resultSetConcurrency);
    }


    /**
     * Creates a CallableStatement for the given SQL, result set type and concurency
     */
    public CallableStatement prepareCall(String sql, int resultSetType, int resultSetConcurrency) throws SQLException {
        return new StandardXACallableStatement(this, sql, resultSetType, resultSetConcurrency);
    }

    /**
     * Creates a CallableStatement for the given SQL
     */
    public CallableStatement prepareCall(String sql) throws SQLException {
        return new StandardXACallableStatement(this, sql, 0, 0);
    }

    public Statement createStatement() throws SQLException {
        if (tx == null) {
            log.debug("StandardXAConnectionHandle:createStatement tx==null");
             try {
                 try {
                     Transaction ntx = (transactionManager!=null)?transactionManager.getTransaction():null;
                     if (ntx != null) {
                         tx = ntx;
                         this.setAutoCommit(false);
                         try {
                             tx.enlistResource(xacon.getXAResource());// enlist the xaResource in the transaction
                         } catch (RollbackException n) {
                             throw new SQLException("StandardXAConnectionHandle:createStatement enlistResource exception: "+n.toString());
                         }
                     }

                 } catch (SystemException n) {
                     throw new SQLException("StandardXAConnectionHandle:createStatement getTransaction exception: "+n.toString());
                 }
             } catch (NullPointerException n) {
                 // current is null: we are not in EJBServer.
                 throw new SQLException("StandardXAConnectionHandle:createStatement should not be used outside an EJBServer: "+n.toString());
             }
        }
        isReallyUsed = true;
        return new StandardXAStatement(this, 0, 0);
    }

    public Statement createStatement(int resultSetType, int resultSetConcurrency) throws SQLException {
        return new StandardXAStatement(this, resultSetType, resultSetConcurrency);
    }
}
